from unittest.mock import Mock, patch
import pytest
import torch
import os
import sys
import warnings

warnings.filterwarnings('ignore')
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from botorch.utils.sampling import manual_seed
from qpots.acquisition import Acquisition
from qpots.function import Function
from qpots.model_object import ModelObject

@pytest.fixture
def mock_gp_no_cons():
    gps = Mock()
    gps.nobj = 2
    gps.ncons = 0
    gps.bounds = torch.tensor([[0.0, 0.0], [1.0, 1.0]])
    gps.train_x = torch.tensor([[0.0, 0.0], [1.0, 1.0]])
    gps.train_y = torch.tensor([[1.0, 2.0], [2.0, 1.0]])
    return gps

@pytest.fixture()
def mock_gp_cons():
    gps = Mock()
    gps.nobj = 2
    gps.ncons = 1
    gps.bounds = torch.tensor([[0.0, 0.0], [1.0, 1.0]])
    gps.train_x = torch.tensor([[1.0, 2.0], [2.0, 4.0]])
    gps.train_y = torch.tensor([[1.0, 5.0, -0.5], [2.0, 3.0, 0.3]])
    return gps

@pytest.fixture
def mock_func():
    func = Mock()
    func.name = "mock_function"
    func.dim = 2
    return func

@pytest.fixture
def real_func():
    func = Function('branincurrin', dim=2, nobj=2)
    return func

def test_acquisition_init(mock_gp_cons, mock_func):
    cons = Mock()

    # Parameters
    device = torch.device("cpu")
    q = 5
    num_restarts = 20
    raw_samples = 1024

    # Instantiate Acquisition
    acquisition = Acquisition(
        func=mock_func,
        gps=mock_gp_cons,
        cons=cons,
        device=device,
        q=q,
        NUM_RESTARTS=num_restarts,
        RAW_SAMPLES=raw_samples,
    )

    # Assertions
    assert acquisition.func == mock_func
    assert acquisition.gps == mock_gp_cons
    assert acquisition.cons == cons
    assert acquisition.device == device
    assert acquisition.q == q
    assert acquisition.NUM_RESTARTS == num_restarts
    assert acquisition.RAW_SAMPLES == raw_samples
    assert acquisition.nobj == mock_gp_cons.nobj
    assert acquisition.ncons == mock_gp_cons.ncons

def test_gp_posterior(mock_gp_no_cons, mock_func):
    # Mock GP models
    model_1 = Mock()
    model_1.posterior.return_value.sample.return_value = torch.tensor([[1.5], [1.6]])
    model_2 = Mock()
    model_2.posterior.return_value.sample.return_value = torch.tensor([[2.5], [2.6]])
    mock_gp_no_cons.models = [model_1, model_2]

    # Create Acquisition instance
    acquisition = Acquisition(func=mock_func, gps=mock_gp_no_cons)
    x = torch.tensor([[0.5, 0.5], [0.3, 0.7]])
    result = acquisition.gp_posterior(x, mock_gp_no_cons)

    # Assertions
    assert isinstance(result, torch.Tensor), "Output is not a tensor"
    assert result.shape == torch.Size([2, 2]), "Output shape is incorrect"
    model_1.posterior.assert_called_once()
    model_2.posterior.assert_called_once()

def test_gp_posterior_with_constraints():
    # Mock inputs
    func = Mock()
    func.dim = 2

    gps = Mock()
    gps.nobj = 2
    gps.ncons = 1  # Adding one constraint
    gps.bounds = torch.tensor([[0.0, 0.0], [1.0, 1.0]])
    
    # Objectives + constraints (nobj + ncons columns)
    gps.train_y = torch.tensor([
        [1.0, 2.0, -0.5],  # Infeasible point (constraint violated)
        [2.0, 1.0, 0.3]    # Feasible point
    ])

    # Mock GP models
    model_1 = Mock()
    model_1.posterior.return_value.sample.return_value = torch.tensor([[1.5, 2.6]])
    model_2 = Mock()
    model_2.posterior.return_value.sample.return_value = torch.tensor([[2.5, 3.6]])
    model_3 = Mock()
    model_3.posterior.return_value.sample.return_value = torch.tensor([[-1.0, 0.5]])
    gps.models = [model_1, model_2, model_3]

    # Create Acquisition instance
    acquisition = Acquisition(func=func, gps=gps)

    # Input tensor
    x = torch.tensor([[0.5, 0.5], [0.3, 0.7]])

    # Test gp_posterior
    result = acquisition.gp_posterior(x, gps)

    # Assertions
    assert isinstance(result, torch.Tensor), "Output is not a tensor"
    assert result.shape == torch.Size([2, 2]), f"Output shape is incorrect: {result.shape}"

    # Check that infeasible points are penalized
    assert torch.all(result[0, :] == 1e12), "Infeasible point not penalized"
    assert torch.any(result[1, :] != 1e12), "Feasible point incorrectly penalized"

    # Verify that posterior calls were made
    model_1.posterior.assert_called_once()
    model_2.posterior.assert_called_once()
    model_3.posterior.assert_called_once()

@patch("qpots.utils.pymoo_problem.nsga2")
@pytest.mark.parametrize("q", [1, 2, 3, 4])
def test_qpots(mock_func, real_func, q):
    pop_size = 100

    train_x = torch.rand(pop_size, real_func.dim, dtype=torch.float64)
    train_y = real_func.evaluate(train_x)
    gps = ModelObject(train_x=train_x, train_y=train_y, bounds=real_func.get_bounds(), nobj=real_func.nobj, ncons=0, device=torch.device("cpu"))
    gps.fit_gp()
    acq = Acquisition(func=mock_func, gps=gps, q=q)

    bounds = torch.tensor([[0.0, 0.0], [1.0, 1.0]])
    iteration = 1
    kwargs = {
        "nystrom": 0,
        "iters": 10,
        "dim": 2,
        "nychoice": "random",
        "q": q,
        "ngen": 10,
    }
    res = acq.qpots(bounds, iteration, **kwargs)

    assert isinstance(res, torch.Tensor), "Output is not a Tensor"
    assert res.shape == torch.Size([q, 2])

@pytest.mark.parametrize("q", [1, 2, 3, 4])
def test_qpots_nystrom(real_func, q):
    pop_size = 100

    train_x = torch.rand(pop_size, real_func.dim, dtype=torch.float64)
    train_y = real_func.evaluate(train_x)
    gps = ModelObject(train_x=train_x, train_y=train_y, bounds=real_func.get_bounds(), nobj=real_func.nobj, ncons=0, device=torch.device("cpu"))
    gps.fit_gp()
    acq = Acquisition(func=mock_func, gps=gps, q=q)

    bounds = torch.tensor([[0.0, 0.0], [1.0, 1.0]])
    iteration = 1
    kwargs = {
        "nystrom": 1,
        "iters": 10,
        "dim": 2,
        "nychoice": "random",
        "q": q,
        "ngen": 10,
    }
    res = acq.qpots(bounds, iteration, **kwargs)

    assert isinstance(res, torch.Tensor), "Output is not a Tensor"
    assert res.shape == torch.Size([q, 2])    

@pytest.mark.parametrize("q", [1, 2, 3, 4])
def test_qlogei(real_func, q):
    train_x = torch.rand(5, real_func.dim, dtype=torch.float64)
    train_y = real_func.evaluate(train_x)

    # Create GPs for each objective
    gps = ModelObject(train_x=train_x, train_y=train_y, bounds=real_func.get_bounds(), nobj=real_func.nobj, ncons=0, device=torch.device("cpu"))
    gps.fit_gp()

    # Create Acquisition instance
    acquisition = Acquisition(func=real_func, gps=gps, q=q)

    # Reference point for hypervolume calculation
    ref_point = torch.tensor([-300.0, -18.0])

    # Call qlogei
    manual_seed(42)# Fix seed for reproducibility
    result = acquisition.qlogei(ref_point)

    # Assertions
    assert isinstance(result, torch.Tensor), "Output is not a tensor"
    assert result.shape == torch.Size([q, 2]), f"Output shape is incorrect: {result.shape}"
    assert (result >= 0.0).all() and (result <= 1.0).all(), "Result is out of bounds"

@pytest.mark.parametrize("q", [1, 2, 3, 4])
def test_parego(real_func, q):
    train_x = torch.rand(5, real_func.dim, dtype=torch.float64)
    train_y = real_func.evaluate(train_x)

    gps = ModelObject(train_x=train_x, train_y=train_y, bounds=real_func.get_bounds(), nobj=real_func.nobj, ncons=0, device=torch.device("cpu"))
    gps.fit_gp()

    acquisition = Acquisition(func=real_func, gps=gps, q=q)

    manual_seed(42)
    result = acquisition.parego()

    assert isinstance(result, torch.Tensor), "Output is not a tensor"
    assert result.shape == torch.Size([q, 2]), f"Output shape is incorrect: {result.shape}"
    assert (result >= 0.0).all() and (result <= 1.0).all(), "Result is out of bounds"

@pytest.mark.parametrize("q", [1, 2, 3, 4])
def test_qlogparego(real_func, q):
    train_x = torch.rand(5, real_func.dim, dtype=torch.float64)
    train_y = real_func.evaluate(train_x)

    gps = ModelObject(train_x=train_x, train_y=train_y, bounds=real_func.get_bounds(), nobj=real_func.nobj, ncons=0, device=torch.device("cpu"))
    gps.fit_gp()

    acquisition = Acquisition(func=real_func, gps=gps, q=q)

    manual_seed(42)
    result = acquisition.qlogparego()

    assert isinstance(result, torch.Tensor), "Output is not a tensor"
    assert result.shape == torch.Size([q, 2]), f"Output shape is incorrect: {result.shape}"
    assert (result >= 0.0).all() and (result <= 1.0).all(), "Result is out of bounds"

@pytest.mark.parametrize("q", [1, 2, 3, 4])
def test_pesmo(real_func, q):
    train_x = torch.rand(100, real_func.dim, dtype=torch.float64)
    train_y = real_func.evaluate(train_x)

    gps = ModelObject(train_x=train_x, train_y=train_y, bounds=real_func.get_bounds(), nobj=real_func.nobj, ncons=0, device=torch.device("cpu"))
    gps.fit_gp()

    acquisition = Acquisition(func=real_func, gps=gps, q=q)

    manual_seed(42)
    result = acquisition.pesmo()

    assert isinstance(result, torch.Tensor), "Output is not a tensor"
    assert result.shape == torch.Size([q, 2]), f"Output shape is incorrect: {result.shape}"
    assert (result >= 0.0).all() and (result <= 1.0).all(), "Result is out of bounds"

@pytest.mark.parametrize("q", [1, 2, 3, 4])
def test_mesmo(real_func, q):
    train_x = torch.rand(100, real_func.dim, dtype=torch.float64)
    train_y = real_func.evaluate(train_x)

    gps = ModelObject(train_x=train_x, train_y=train_y, bounds=real_func.get_bounds(), nobj=real_func.nobj, ncons=0, device=torch.device("cpu"))
    gps.fit_gp()

    acquisition = Acquisition(func=real_func, gps=gps, q=q)

    manual_seed(42)
    result = acquisition.mesmo()

    assert isinstance(result, torch.Tensor), "Output is not a tensor"
    assert result.shape == torch.Size([q, 2]), f"Output shape is incorrect: {result.shape}"
    assert (result >= 0.0).all() and (result <= 1.0).all(), "Result is out of bounds"

@pytest.mark.parametrize("q", [1, 2, 3])
def test_jesmo(real_func, q):
    train_x = torch.rand(100, real_func.dim, dtype=torch.float64)
    train_y = real_func.evaluate(train_x)

    gps = ModelObject(train_x=train_x, train_y=train_y, bounds=real_func.get_bounds(), nobj=real_func.nobj, ncons=0, device=torch.device("cpu"))
    gps.fit_gp_no_variance()

    acquisition = Acquisition(func=real_func, gps=gps, q=q)

    manual_seed(42)
    result = acquisition.jesmo()

    assert isinstance(result, torch.Tensor), "Output is not a tensor"
    assert result.shape == torch.Size([q, 2]), f"Output shape is incorrect: {result.shape}"
    assert (result >= 0.0).all() and (result <= 1.0).all(), "Result is out of bounds"



import torch
from botorch.utils.multi_objective.box_decompositions import FastNondominatedPartitioning
from botorch.utils.multi_objective.hypervolume import Hypervolume
from botorch.utils.multi_objective.hypervolume import Hypervolume
from botorch.utils.multi_objective.pareto import is_non_dominated
from sklearn.neighbors import KernelDensity
from scipy.spatial.distance import cdist
import argparse


def unstandardize(Y, train_y):
    """
    Reverse the standardization process for the output Y using the mean and standard deviation
    from the training data.

    Parameters:
        Y (torch.Tensor): The standardized output tensor.
        train_y (torch.Tensor): The training output data used to compute the mean and std dev.

    Returns:
        torch.Tensor: The unstandardized output tensor.
    """
    mean = train_y.mean(dim=0)
    std = train_y.std(dim=0)
    return Y * std + mean

def expected_hypervolume(gps, ref_point=torch.tensor([-300., -18.]), min=False):
    """
    Compute the expected hypervolume and Pareto front based on the GP model predictions.

    Parameters:
        gps: The multi-objective GP models.
        ref_point (torch.Tensor): Reference point for hypervolume calculation.

    Returns:
        tuple: (Hypervolume value, Pareto front tensor).
    """
    # Compute hypervolume using Pareto front with feasible points if constraints exist
    if min:
        if gps.ncons > 0:
            is_feas = (gps.train_y[..., -gps.ncons:] >= 0).all(dim=-1)
            is_feas_obj = gps.train_y[is_feas]
            pareto_mask = is_non_dominated(is_feas_obj, maximize=False) 
            pareto_front = is_feas_obj[pareto_mask]
            hv_calculator = Hypervolume(ref_point=-1*torch.tensor([0.335, 0.335]))
            hypervolume_value = hv_calculator.compute(-1*pareto_front)
            return hypervolume_value, pareto_front
        else:
            pareto_mask = is_non_dominated(gps.train_y, maximize=False) 
            pareto_front = gps.train_y[pareto_mask]
            hv_calculator = Hypervolume(ref_point=-1*torch.tensor([0.335, 0.335]))
            hypervolume_value = hv_calculator.compute(-1*pareto_front)
            return hypervolume_value, pareto_front
    else:
        if gps.ncons > 0:
            is_feas = (gps.train_y[..., -gps.ncons:] >= 0).all(dim=-1)
            is_feas_obj = gps.train_y[is_feas]
            bd1 = FastNondominatedPartitioning(ref_point.double(), is_feas_obj.double()[..., :gps.nobj])
            return bd1.compute_hypervolume(), bd1.pareto_Y
        else:
            bd1 = FastNondominatedPartitioning(ref_point.double(), gps.train_y[..., :gps.nobj].double())
            return bd1.compute_hypervolume(), bd1.pareto_Y

def gen_filtered_cands(gps, cands, ref_point=torch.tensor([0., 0.]), kernel_bandwidth=0.05):
    """
    Generate filtered candidate points based on the current Pareto front using Kernel Density Estimation (KDE).

    Parameters:
        gps: The multi-objective GP models.
        cands (torch.Tensor): Candidate points to filter.
        ref_point (torch.Tensor): Reference point for the Pareto front.
        kernel_bandwidth (float): Bandwidth for the KDE filter.

    Returns:
        torch.Tensor: Filtered candidate points.
    """
    # Compute the Pareto front based on the training data
    bd1 = FastNondominatedPartitioning(ref_point.double(), gps.train_y)
    nPareto = bd1.pareto_Y.shape[0]

    # Find the indices of Pareto optimal points in the training data
    ind = torch.tensor([(gps.train_y == bd1.pareto_Y[j]).nonzero()[0, 0] for j in range(nPareto)])
    x_nd = gps.train_x[ind]

    # Fit Kernel Density Estimation (KDE) using the Pareto optimal points
    kde = KernelDensity(kernel='gaussian', bandwidth=kernel_bandwidth).fit(x_nd)

    # Filter candidates based on the KDE
    U = torch.log(torch.rand(cands.shape[0]))
    w = kde.score_samples(cands)
    M = w.max()
    cands_fil = cands[w > U.numpy() * M]
    
    return cands_fil
    
def select_candidates(gps, pareto_set, device, q=1, seed=None):
        """
        Select candidates from the Pareto set.

        Parameters:
            gps: Gaussian Process models.
            pareto_set: Pareto optimal set of solutions.
            k: Number of samples to select.
            seed: Random seed for sampling.

        Returns:
            torch.Tensor: Selected candidates.
        """
        if seed is not None:
            torch.manual_seed(seed)

        D = cdist(pareto_set, gps.train_x.numpy())
        selected_indices = D.min(axis=-1).argsort()[-q:]
        selected_candidates = torch.from_numpy(pareto_set[selected_indices]).to(torch.double).to(device)
        return selected_candidates

def arg_parser():
    """
    This function contains all the necessary command line arguments. Each one has a default so it is not always necessary to specify them
    unless the user desires a different value of the argument.

    Arguments marked HPC are helpful for an HPC environment
    """
    # Argument parsing
    parser = argparse.ArgumentParser(description="Multi-objective Bayesian Optimization")
    parser.add_argument('--ntrain', type=int, default=20, help='Number of initial points')
    parser.add_argument('--iters', type=int, default=20, help='Number of iterations')
    parser.add_argument('--reps', type=int, default=20, help='Number of repetitions')
    parser.add_argument('--q', type=int, default=1, help='Number of points to sample at each iteration')
    parser.add_argument('--wd', type=str, default=".", help='Working directory to save results')
    parser.add_argument('--func', type=str, default=None, help='Test function to optimize. If using a custom function then don not specify') # HPC
    parser.add_argument('--ref_point', type=float, nargs='+', required=True, help='Reference point for hypervolume calculation') # HPC
    parser.add_argument('--dim', type=int, required=True, help='Input dimensionality')
    parser.add_argument('--nobj', type=int, default=2, help='Number of objectives')
    parser.add_argument('--ncons',type=int, default=0, help='Number of constraints')
    parser.add_argument('--acq', type=str, default="TS", help='Acquisition function') # HPC
    parser.add_argument('--nystrom', type=int, default=0, help='Use Nystrom approximation with filtered candidates')
    parser.add_argument('--nychoice', type=str, default='pareto', help='Pick a nystrom choice of m, pareto or random')
    parser.add_argument('--ngen', type=int, default=10, help='Number of generations for NSGA-II')
    args = parser.parse_args()

    return args



from unittest.mock import Mock
import pytest
import torch
import os
import sys
import warnings

warnings.filterwarnings('ignore')
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from botorch.test_functions.multi_objective import BraninCurrin, OSY
from qpots.function import Function

@pytest.fixture()
def mock_func():
    func = Mock()
    func.cons = Mock()
    func.bounds = Mock()
    func.name = 'branincurrin'
    func.dim = 2
    func.nobj = 2
    return func

@pytest.fixture()
def custom_function():
    def custom_func(X):
        f1 = X[:, 0]**2 + X[:, 1]**2 + 2*X[:, 0]*X[:, 1]**2
        f2 = (X[:, 0] - 1)**2 + (X[:, 1] - 1)**2
        return -1*torch.stack([f1, f2], dim=-1)
    
    return custom_func

def test_function_init(mock_func):
    name = mock_func.name 
    dim = mock_func.dim
    nobj = mock_func.nobj
    custom_func = mock_func.custom_func
    bounds = mock_func.bounds
    cons = mock_func.cons

    f = Function(
        name=name,
        dim=dim,
        nobj=nobj,
        custom_func=custom_func,
        bounds=bounds,
        cons=cons
    )

    assert f.name == name
    assert f.dim == dim
    assert f.nobj == nobj
    assert f.custom_func == custom_func
    assert f.bounds == bounds
    assert f.cons == cons

def test_func_map_properly_initialized():
    obj = Function('branincurrin', dim=2, nobj=2)
    assert isinstance(obj.func, BraninCurrin), "Function name does not match function instance"
    assert obj.bounds.shape == (2,2), "Bounds are not correct shape"

def test_func_map_with_cons():
    obj = Function('osy')
    assert isinstance(obj.func, OSY), "Function name does not match function instance"
    assert obj.bounds.shape == (2, 6), "Bound are not correct shape"
    assert hasattr(obj, 'cons'), "Cons not initialized"

def test_func_map_evaluate():
    obj = Function('branincurrin', dim=2, nobj=2)
    x = torch.tensor([[0.2, 0.4]])
    res = obj.evaluate(x)
    assert isinstance(res, torch.Tensor), "Result is not a Tensor object"
    assert res.shape == torch.Size([1, obj.dim]), "Shape does not match expected shape"

def test_func_map_with_cons():
    obj = Function('osy', dim=6, nobj=2)
    cons = obj.get_cons()
    x = torch.tensor([[0.2, 0.4, 0.3, 0.2, 0.5, 0.6]])
    res = obj.evaluate(x)
    assert isinstance(res, torch.Tensor), "Result is not a Tensor object"
    assert res.shape == torch.Size([1, obj.nobj]), "Shape does not match expected shape"
    assert isinstance(cons(x), torch.Tensor), "Constraint does not return a Tensor object"
    assert cons(x).shape == torch.Size([1, 6]) # This problem has 6 constraints so 6 "constraint outputs"

def test_custom_function(custom_function):
    x = torch.tensor([[0.2, 0.4]])
    bounds = torch.tensor(([0.0, 0.0], [1.0, 1.0]))
    cons = None
    obj = Function(name=None, dim=2, nobj=2, custom_func=custom_function, bounds=bounds, cons=cons)
    assert isinstance(obj.evaluate(x), torch.Tensor), "Result not a Tensor"
    assert obj.evaluate(x).shape == torch.Size([1, obj.nobj]), "Shapes do not match"

def test_custom_function_with_cons(custom_function):
    x = torch.tensor([0.2, 0.4])
    bounds = torch.tensor(([0.0, 0.0], [1.0, 1.0]))
    def cons(x):
        con1 = x[0] + x[1]*x[0]
        return -1*torch.stack([con1], dim=-1)
    obj = Function(name=None, dim=2, nobj=2, custom_func=custom_function, bounds=bounds, cons=cons)
    cons_func = obj.get_cons()
    assert isinstance(cons_func(x), torch.Tensor), "Tensor not returned from custom constraint function"
    assert cons_func(x).shape == torch.Size([1]), "Size is not properly returned"

def test_custom_function_missing_bounds(custom_function):
    with pytest.raises(ValueError, match="Custom functions must specify bounds."):
        Function(name=None, dim=2, nobj=2, custom_func=custom_function, bounds=None, cons=None)

def test_invalid_function_name():
    with pytest.raises(ValueError, match="Unknown test function 'invalid_name'."):
        Function(name="invalid_name", dim=2, nobj=2)

def test_get_bounds():
    obj = Function('branincurrin', dim=2, nobj=2)
    assert isinstance(obj.get_bounds(), torch.Tensor), "Returned bounds are not a Tensor object"
    assert torch.equal(obj.get_bounds(), obj.bounds), "get_bounds() does not return the correct bounds"

def test_get_bounds_with_custom_function(custom_function):
    bounds = torch.tensor([[0.0, 0.0], [1.0, 1.0]])
    obj = Function(dim=2, nobj=2, custom_func=custom_function, bounds=bounds, cons=None)
    assert isinstance(obj.get_bounds(), torch.Tensor), "Returned bounds are not a Tensor object"
    assert torch.equal(obj.get_bounds(), obj.bounds), "get_bounds() does not return the correct bounds"

def test_get_cons_without_constraints():
    obj = Function('branincurrin', dim=2, nobj=2)
    assert obj.get_cons() is None, "get_cons() should return None for unconstrained functions"

def test_func_map_batch_evaluate():
    obj = Function('branincurrin', dim=2, nobj=2)
    X = torch.tensor([[0.2, 0.4], [0.5, 0.1], [0.9, 0.7]])
    res = obj.evaluate(X)
    assert isinstance(res, torch.Tensor), "Result is not a Tensor"
    assert res.shape == torch.Size([3, obj.nobj]), "Batch evaluation shape mismatch"

def test_custom_function_with_batch_evaluate(custom_function):
    bounds = torch.tensor([[0.0, 0.0], [1.0, 1.0]])
    obj = Function(dim=2, nobj=2, custom_func=custom_function, bounds=bounds, cons=None)
    X = torch.tensor([[0.2, 0.4], [0.5, 0.1], [0.9, 0.7]])
    res = obj.evaluate(X)
    assert isinstance(res, torch.Tensor), "Result is not a Tensor"
    assert res.shape == torch.Size([3, obj.nobj]), "Batch evaluation shape mismatch"
    


from unittest.mock import Mock
import pytest
import torch
import numpy as np
import os
import sys
import warnings

warnings.filterwarnings('ignore')
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from botorch.utils.multi_objective.box_decompositions import FastNondominatedPartitioning
from botorch.utils.multi_objective.hypervolume import Hypervolume

from qpots.utils.utils import unstandardize, expected_hypervolume, gen_filtered_cands, select_candidates, arg_parser

@pytest.fixture
def mock_gps():
    gps = Mock()
    gps.train_y = torch.tensor([[1.0, 2.0], [2.0, 1.0], [1.5, 1.5]])
    gps.train_x = torch.tensor([[0.1, 0.2], [0.2, 0.1], [0.15, 0.15]])
    gps.nobj = 2
    gps.ncons = 0
    return gps

def test_unstandardize():
    Y = torch.tensor([[1.0, 2.0], [3.0, 4.0]])
    train_y = torch.tensor([[2.0, 3.0], [4.0, 5.0]])

    result = unstandardize(Y, train_y)
    mean = train_y.mean(dim=0)
    std = train_y.std(dim=0)
    expected = Y * std + mean

    assert torch.allclose(result, expected), "Unstandardization failed"

def test_expected_hypervolume(mock_gps):
    ref_point = torch.tensor([-1.0, -1.0])

    hv_value, pareto_front = expected_hypervolume(mock_gps, ref_point)
    
    # Verify results
    hv = Hypervolume(ref_point=ref_point.double())
    partitioning = FastNondominatedPartitioning(ref_point.double(), mock_gps.train_y.double())
    expected_pareto = partitioning.pareto_Y
    expected_hv = hv.compute(expected_pareto)

    assert torch.allclose(hv_value, torch.tensor([expected_hv], dtype=torch.double)), "Hypervolume computation failed"
    assert torch.equal(pareto_front, expected_pareto), "Pareto front mismatch"

def test_gen_filtered_cands(mock_gps):
    candidates = torch.tensor([[0.1, 0.2], [0.3, 0.4], [0.5, 0.6]])
    ref_point = torch.tensor([0.0, 0.0])

    filtered_candidates = gen_filtered_cands(mock_gps, candidates, ref_point, kernel_bandwidth=0.05)
    
    assert filtered_candidates.shape[0] <= candidates.shape[0], "Too many candidates selected"

def test_select_candidates(mock_gps):
    pareto_set = np.array([[0.1, 0.1], [0.9, 0.9]])
    device = "cpu"

    selected_candidates = select_candidates(mock_gps, pareto_set, device, q=1, seed=42)
    assert selected_candidates.shape[0] == 1, "Incorrect number of candidates selected"

def test_arg_parser():
    sys.argv = [
        "script_name",
        "--ntrain", "10",
        "--ref_point", "0", "0",
        "--dim", "3"
    ]
    args = arg_parser()
    assert args.ntrain == 10, "Failed to parse --ntrain"
    assert args.dim == 3, "Failed to parse --dim"
    assert args.ref_point == [0.0, 0.0], "Failed to parse --ref_point"



from botorch.test_functions.multi_objective import (
    BraninCurrin, DTLZ1, DTLZ2, DTLZ3, DTLZ7, GMM, DH1, Penicillin,
    VehicleSafety, CarSideImpact, ConstrainedBraninCurrin,
    ZDT3, DiscBrake, MW7, OSY, WeldedBeam
)
from torch import Tensor
from typing import Callable, Optional, Union

class Function:
    """
    Interface for multi-objective test functions.
    Provides an easy-to-use abstraction over BoTorch test functions and supports user-defined functions.
    """
    def __init__(
        self,
        name: Optional[str] = None,
        dim: int = 2,
        nobj: int = 2,
        custom_func: Optional[Callable[[Tensor], Tensor]] = None,
        bounds: Optional[Tensor] = None,
        cons: Optional[Callable[[Tensor], Tensor]] = None,
    ):
        """
        Args:
            name (str, optional): Name of the predefined test function (case-insensitive).
            dim (int): Dimensionality of the input space.
            nobj (int): Number of objectives for the test function.
            custom_func (Callable, optional): User-defined function for optimization.
            bounds (Tensor, optional): Bounds for the custom function.
            cons (Callable, optional): Constraint function for the custom function.
        """
        self.name = name.lower() if name else None
        self.dim = dim
        self.nobj = nobj
        self.custom_func = custom_func
        self.bounds = bounds
        self.cons = cons

        if self.custom_func:
            # Validate custom function setup
            if self.bounds is None:
                raise ValueError("Custom functions must specify bounds.")
        else:
            # Initialize predefined test functions
            self._initialize_function()

    def _initialize_function(self):
        # Mapping between function names and BoTorch test functions
        func_map = {
            "branincurrin": lambda: BraninCurrin(negate=True),
            "dtlz1": lambda: DTLZ1(self.dim, num_objectives=self.nobj, negate=True),
            "dtlz2": lambda: DTLZ2(self.dim, num_objectives=self.nobj, negate=True),
            "dtlz3": lambda: DTLZ3(self.dim, num_objectives=self.nobj),
            "dtlz7": lambda: DTLZ7(self.dim, num_objectives=self.nobj, negate=True),
            "dh1": lambda: DH1(self.dim, negate=True),
            "gmm": lambda: GMM(self.nobj, negate=True),
            "penicillin": lambda: Penicillin(negate=True),
            "vehicle": lambda: VehicleSafety(negate=True),
            "carside": lambda: CarSideImpact(negate=True),
            "zdt3": lambda: ZDT3(dim=self.dim, num_objectives=self.nobj, negate=True),
            "constrainedbc": lambda: ConstrainedBraninCurrin(negate=True),
            "discbrake": lambda: DiscBrake(),
            "mw7": lambda: MW7(dim=self.dim, negate=True),
            "osy": lambda: OSY(negate=True),
            "weldedbeam": lambda: WeldedBeam(negate=True),
        }

        if self.name not in func_map:
            raise ValueError(f"Unknown test function '{self.name}'. Check the name or documentation.")

        # Initialize the function, bounds, and constraints
        self.func = func_map[self.name]()
        self.bounds = self.func.bounds.double()
        if hasattr(self.func, "evaluate_slack"):
            self.cons = self.func.evaluate_slack

    def evaluate(self, X: Tensor) -> Tensor:
        """
        Evaluate the test function or custom function on input X.

        Args:
            X (Tensor): Input tensor of shape (n, dim).

        Returns:
            Tensor: Outputs from the function.
        """
        if self.custom_func:
            return self.custom_func(X)
        return self.func(X)

    def get_bounds(self) -> Tensor:
        """
        Retrieve the bounds for the function.

        Returns:
            Tensor: Lower and upper bounds as a tensor.
        """
        return self.bounds

    def get_cons(self) -> Optional[Callable]:
        """
        Retrieve the constraints for the function.

        Returns:
            Callable or None: Constraint function if available.
        """
        return self.cons
  


import pytest
import torch
import numpy as np
import os
from unittest.mock import Mock, patch, MagicMock
import warnings
import sys
warnings.filterwarnings('ignore')
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from qpots.tsemo_runner import TSEMORunner

@pytest.fixture
def tsemo_runner():
    with patch("matlab.engine.start_matlab", return_value=Mock()) as mock_engine:
        runner = TSEMORunner(
            func="test_function",
            x=torch.tensor([[0.1, 0.2], [0.3, 0.4]]),
            y=torch.tensor([[1.0, 2.0], [3.0, 4.0]]),
            lb=[0, 0],
            ub=[1, 1],
            iters=5,
            batch_number=2
        )
        return runner

def test_tsemo_runner_init(tsemo_runner):
    assert tsemo_runner._func == "test_function"
    assert len(tsemo_runner._x) == 2
    assert len(tsemo_runner._y) == 2
    assert tsemo_runner._iters == 5
    assert tsemo_runner._batch_number == 2
    assert isinstance(tsemo_runner._eng, Mock)  # MATLAB engine should be mocked

def test_tsemo_run(tsemo_runner, tmp_path):
    mock_X_out = [[0.5, 0.6], [0.7, 0.8]]
    mock_Y_out = [[5.0, 6.0], [7.0, 8.0]]
    mock_times = [0.1, 0.2]

    tsemo_runner._eng.TSEMO_run = MagicMock(return_value=(mock_X_out, mock_Y_out, mock_times))

    save_dir = tmp_path  # Temporary directory for testing file saving
    rep = 1

    X, Y, times_np = tsemo_runner.tsemo_run(save_dir, rep)

    # Ensure the MATLAB function was called correctly
    tsemo_runner._eng.TSEMO_run.assert_called_once()

    # Verify returned values
    assert isinstance(X, np.ndarray)
    assert isinstance(Y, np.ndarray)
    assert isinstance(times_np, np.ndarray)
    assert X.shape == (2, 2)
    assert Y.shape == (2, 2)

    # Check that files were saved correctly
    assert (save_dir / "X_1.npy").exists()
    assert (save_dir / "Y_1.npy").exists()
    assert (save_dir / "times_1.npy").exists()

    # Check file contents
    np.testing.assert_array_equal(np.load(save_dir / "X_1.npy"), np.array(mock_X_out))
    np.testing.assert_array_equal(np.load(save_dir / "Y_1.npy"), np.array(mock_Y_out))
    np.testing.assert_array_equal(np.load(save_dir / "times_1.npy"), np.array(mock_times))

def test_tsemo_hypervolume():
    Y = torch.tensor([[1.0, 2.0], [2.0, 1.5], [1.5, 1.8], [3.0, 2.5]])
    ref_point = torch.tensor([4.0, 4.0])
    train_shape = 2
    iters = 2

    tsemo_runner = TSEMORunner(
        func="test_function",
        x=[],
        y=[],
        lb=[],
        ub=[],
        iters=iters,
        batch_number=1
    )

    hv, pf = tsemo_runner.tsemo_hypervolume(Y, ref_point, train_shape, iters)

    assert isinstance(hv, list), "Hypervolume should be a list"
    assert len(hv) == iters, "Hypervolume list length mismatch"
    assert isinstance(pf, torch.Tensor), "Pareto front should be a tensor"
    assert pf.shape[1] == 2, "Pareto front shape mismatch"



import torch
from botorch.models import SingleTaskGP
from gpytorch.mlls.exact_marginal_log_likelihood import ExactMarginalLogLikelihood
from botorch.fit import fit_gpytorch_mll
from botorch.utils.transforms import standardize

class ModelObject:
    """
    A class to represent multi-objective GP models using independent MLE Gaussian Processes (GPs) 
    for each objective. Each GP is trained using maximum likelihood estimation (MLE) with fixed noise.
    """

    def __init__(self, train_x, train_y, bounds,
                 nobj,
                 ncons,
                 device, 
                 noise_std=1e-2):
        """
        Initialize the multi-objective GP models.

        Parameters:
            train_x (torch.Tensor): Training input data.
            train_y (torch.Tensor): Training output data (multi-objective).
            bounds (torch.Tensor): Bounds for the input data.
            nobj (int): Number of objective functions.
            ncons (int): Number of constraints in the problem.
            device (str): CUDA or CPU device for the arrays to operate on
            noise_std (float): Standard deviation of noise for the GP model. Default is 1e-2.
        """
        self.train_x = train_x.to(device)
        self.train_y = train_y.to(device)
        self.noise_std = noise_std
        self.bounds = bounds
        self.nobj = nobj
        self.ncons= ncons
        self.device = device
        self.models = []
        self.mlls = []

    def fit_gp(self):
        """
        Fit Gaussian Process (GP) models for each objective using maximum likelihood estimation (MLE).
        The method creates K independent GP models, where K is the number of objectives, and fits them 
        using exact marginal log likelihood.

        Returns:
            None
        """
        num_outputs = self.train_y.shape[-1]
        print("fitting GPs", flush=True)
        train_yvar = torch.ones_like(self.train_y[..., 0], dtype=torch.double).to(self.device).reshape(-1, 1) * self.noise_std ** 2

        # Loop through each objective and fit a GP model
        for i in range(num_outputs):
            print(f"Fit: {i}", flush=True)
            model = SingleTaskGP(
                self.train_x,
                standardize(self.train_y[..., i]).reshape(-1, 1).double(), 
                train_yvar
            ).to(self.train_x.device)
            
            self.models.append(model)
            mll = ExactMarginalLogLikelihood(model.likelihood, model)
            self.mlls.append(mll)

            fit_gpytorch_mll(mll)

    def fit_gp_no_variance(self):
        """
        Fit Gaussian Process (GP) models for each objective using maximum likelihood estimation (MLE).
        The method creates K independent GP models, where K is the number of objectives, and fits them 
        using exact marginal log likelihood. No variance in the SingleTaskGP.

        Returns:
            None
        """
        num_outputs = self.train_y.shape[-1]
        print("fitting GPs", flush=True)

        # Loop through each objective and fit a GP model
        for i in range(num_outputs):
            print(f"Fit: {i}", flush=True)
            model = SingleTaskGP(
                self.train_x,
                standardize(self.train_y[..., i]).reshape(-1, 1).double(), 
            ).to(self.train_x.device)
            
            self.models.append(model)
            mll = ExactMarginalLogLikelihood(model.likelihood, model)
            self.mlls.append(mll)

            fit_gpytorch_mll(mll)


from unittest.mock import Mock
import pytest
import torch
import numpy as np
import os
import sys
import warnings

warnings.filterwarnings('ignore')
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from pymoo.core.result import Result
from qpots.utils.pymoo_problem import PyMooFunction, nsga2

def test_pymoo_problem_init():
    func = Mock()
    n_var = 2
    n_obj = 2
    xl = 0.
    xu = 1.

    problem = PyMooFunction(func=func, n_var=n_var, n_obj=n_obj, xl=xl, xu=xu)

    assert problem.func == func
    assert problem.n_var == n_var
    assert problem.n_obj == n_obj
    assert np.all(problem.xl == xl)
    assert np.all(problem.xu == xu)


def test_nsga2():
    def func(X):
        f1 = X[:, 0]**2 + X[:, 1]**2 + 2*X[:, 0]*X[:, 1]**2
        f2 = (X[:, 0] - 1)**2 + (X[:, 1] - 1)**2
        return -1*torch.stack([f1, f2], dim=-1)
    n_var = 2
    n_obj = 2
    xl = 0.
    xu = 1.

    pop_size = 100
    ngen = 100

    problem = PyMooFunction(func=func, n_var=n_var, n_obj=n_obj, xl=xl, xu=xu)
    res = nsga2(problem=problem, ngen=ngen, pop_size=pop_size, seed=1023, callback=None)
    assert isinstance(res, Result), "res didn't return a Result object"
    assert isinstance(res.X, np.ndarray), "res.X is not an np.ndarray"
    assert isinstance(res.F, np.ndarray), "res.F is not an np.ndarray"
    assert res.X.shape == (pop_size, n_var), "Shape in res.X is mismatched from expected"
    assert res.F.shape == (pop_size, n_obj), "Shape in res.F is mismatched from expected"
    assert np.all(res.X >= xl) and np.all(res.X <= xu), "Out of bounds"

def test_nsga2_with_callback():
    mock_callback = Mock()

    def func(X):
        f1 = X[:, 0]**2 + X[:, 1]**2 + 2*X[:, 0]*X[:, 1]**2
        f2 = (X[:, 0] - 1)**2 + (X[:, 1] - 1)**2
        return -1*torch.stack([f1, f2], dim=-1)
    n_var = 2
    n_obj = 2
    xl = 0.
    xu = 1.

    pop_size = 100
    ngen = 100

    problem = PyMooFunction(func=func, n_var=n_var, n_obj=n_obj, xl=xl, xu=xu)
    res = nsga2(problem=problem, ngen=ngen, pop_size=pop_size, seed=1023, callback=mock_callback)

    assert mock_callback.call_count > 0, "Callback function was never called"
    assert np.all(res.X >= xl) and np.all(res.X <= xu), "Out of bounds"

def test_pymoo_problem_evaluate():
    func = Mock(return_value=torch.tensor([[1.0, 2.0]]))  # Mock function
    problem = PyMooFunction(func=func, n_var=2, n_obj=2, xl=0., xu=1.)

    out = {}
    x_sample = np.array([[0.5, 0.5]])  # Sample input
    problem._evaluate(x_sample, out)

    func.assert_called_once()  # Ensure func is called
    assert "F" in out, "Output dictionary should have 'F'"
    assert out["F"].shape == (1, 2), "Function output shape mismatch"







import pytest
import torch
import os
import sys
import warnings

warnings.filterwarnings('ignore')
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from botorch.models import SingleTaskGP
from gpytorch.mlls.exact_marginal_log_likelihood import ExactMarginalLogLikelihood
from qpots.model_object import ModelObject

@pytest.fixture()
def mock_gp():
    train_x = torch.tensor([[0.4, 0.2]], dtype=torch.float64)
    train_y = torch.tensor([[40., 24.]], dtype=torch.float64)
    bounds = torch.tensor([[0.0, 0.0], [1.0, 1.0]])
    nobj = 2
    ncons = 2
    device = torch.device("cpu")
    noise_std = 1e-6

    gps = ModelObject(
        train_x=train_x, 
        train_y=train_y, 
        bounds=bounds, 
        nobj=nobj, 
        ncons=ncons, 
        device=device,
        noise_std=noise_std)
    
    return gps

def test_model_object_init():
    train_x = torch.tensor([[0.4, 0.2]])
    train_y = torch.tensor([[40., 24.]])
    bounds = torch.tensor([[0.0, 0.0], [1.0, 1.0]])
    nobj, ncons, noise_std = 2, 2, 1e-6
    device = torch.device("cpu")

    gps = ModelObject(train_x, train_y, bounds, nobj, ncons, device, noise_std)

    assert torch.equal(gps.train_x, train_x)
    assert torch.equal(gps.train_y, train_y)
    assert torch.equal(gps.bounds, bounds)
    assert gps.nobj == nobj
    assert gps.ncons == ncons
    assert gps.device == device
    assert gps.noise_std == noise_std

def test_fit_gp(mock_gp):
    mock_gp.fit_gp()
    assert len(mock_gp.models) == mock_gp.nobj
    assert len(mock_gp.mlls) == mock_gp.nobj
    assert all(isinstance(model, SingleTaskGP) for model in mock_gp.models)
    assert all(isinstance(mll, ExactMarginalLogLikelihood) for mll in mock_gp.mlls)

def test_fit_gp_no_variance(mock_gp):
    mock_gp.fit_gp_no_variance()
    assert len(mock_gp.models) == mock_gp.nobj
    assert len(mock_gp.mlls) == mock_gp.nobj
    assert all(isinstance(model, SingleTaskGP) for model in mock_gp.models)
    assert all(isinstance(mll, ExactMarginalLogLikelihood) for mll in mock_gp.mlls)

def test_model_predictions(mock_gp):
    mock_gp.fit_gp()
    test_x = torch.tensor([[0.5, 0.3]], dtype=torch.float64)
    preds = mock_gp.models[0].posterior(test_x)
    assert preds.mean.shape == (1, 1)

def test_no_objectives():
    with pytest.raises(IndexError):
        gps = ModelObject(torch.rand(1, 2), torch.rand(1, 0), torch.rand(2, 2), nobj=0, ncons=0, device="cpu")
        gps.fit_gp()






import torch
from pymoo.core.problem import Problem
from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.optimize import minimize

class PyMooFunction(Problem):
    '''
    Custom test function for use with PyMoo
    '''

    def __init__(self, func, n_var=2, n_obj=2, xl=0., xu=1.):
        self.count = 1
        self.func = func
        self.n_var = n_var
        self.n_obj = n_obj
        self.xl = xl
        self.xu = xu
        super().__init__(n_var=self.n_var,
                         n_obj=self.n_obj,
                         xl=self.xl,
                         xu=self.xu)

    def _evaluate(self, x, out, *args, **kwargs):
        x_ = torch.tensor(x, dtype=torch.double)
        self.count += 1
        out["F"] = self.func(x_).numpy()

def nsga2(problem, ngen=100, pop_size=100, seed=2436, callback=None):
    """
    Perform NSGA-II optimization.

    Parameters:
        problem: Optimization problem object.
        ngen: Number of generations.
        pop_size: Population size.
        seed: Random seed for optimization.

    Returns:
        Optimization result.
    """
    algorithm = NSGA2(pop_size=pop_size)
    if callback:
        res = minimize(problem, algorithm, ('n_gen', ngen), savehistory=True, seed=seed, verbose=False, callback=callback)
    else:
        res = minimize(problem, algorithm, ('n_gen', ngen), savehistory=True, seed=seed, verbose=False)
    return res


import matlab.engine
import numpy as np
import torch
from botorch.utils.multi_objective.box_decompositions import FastNondominatedPartitioning

class TSEMORunner:
    """
    The 'mobo_TSEMO' class runs the TS-EMO algorithm iteratively in MATLAB, 
    updating and saving the results after each iteration.
    """

    def __init__(self, func: str, 
                 x: list, 
                 y: list, 
                 lb: list, 
                 ub: list, 
                 iters: int, 
                 batch_number: int):
        self._func = func
        self._x = x
        self._y = y
        self._lb = lb
        self._ub = ub
        self._iters = iters
        self._batch_number = batch_number
        self._eng = matlab.engine.start_matlab()

        # Add paths to MATLAB environment
        self._eng.addpath(r'./TS-EMO', nargout=0)
        self._eng.addpath(r'./TS-EMO/Test_functions', nargout=0)
        self._eng.addpath(r'./TS-EMO/Direct', nargout=0)
        self._eng.addpath(r'./TS-EMO/Mex_files/invchol', nargout=0)
        self._eng.addpath(r'./TS-EMO/Mex_files/hypervolume', nargout=0)
        self._eng.addpath(r'./TS-EMO/Mex_files/pareto front', nargout=0)
        self._eng.addpath(r'./TS-EMO/NGPM_v1.4', nargout=0)

    def tsemo_run(self, save_dir, rep):
        """
        Run the TS-EMO algorithm iteratively and save results after each iteration.

        Parameters:
            save_dir (str): Directory to save results.
            rep (int): Repetition number to append to the file names.

        Returns:
            None
        """
        # Load the initial X and Y state (assume they are stored in self._x and self._y)
        X = self._x
        Y = self._y
        try:
            [X_out, Y_out, times] = self._eng.TSEMO_run(self._func, 
                                                    matlab.double(X.tolist()), 
                                                    matlab.double(Y.tolist()), 
                                                    matlab.double(self._lb), 
                                                    matlab.double(self._ub), 
                                                    self._iters,
                                                    self._batch_number, 
                                                    nargout=3)
        except matlab.engine.EngineError as e:
            print(f"MATLAB engine encountered an error: {e}")
            raise
        # Convert MATLAB arrays to NumPy arrays
        X_np = np.array(X_out)
        Y_np = np.array(Y_out)
        times_np = np.array(times)

        # Update X and Y with the new outputs from the MATLAB function
        X = X_np
        Y = Y_np

        # Save the updated X, Y, and times after each iteration
        np.save(f"{save_dir}/Y_{rep}.npy", Y_np.squeeze())
        np.save(f"{save_dir}/X_{rep}.npy", X_np)
        np.save(f"{save_dir}/times_{rep}.npy", times_np)

        return X, Y, times_np
    
    def tsemo_hypervolume(self, Y, ref_point, train_shape, iters):
        """
        Calculates the hypervolume and Pareto front of a set of objective values using the Fast Nondominated Partitioning algorithm.
        
        Parameters:
            Y (Tensor): A tensor of objective values, where each row corresponds to an evaluated solution.
            ref_point (Tensor): The reference point used for hypervolume calculation, typically worse than the worst objective values.
            train_shape (int): The number of initial training points, which determines how many points are included in the hypervolume calculation at each step.
            iters (int): Number of iterations that the optimization was ran for.
            
        Returns:
            hv (list): A tensor containing the hypervolume values computed at each iteration.
            pf (Tensor): The Pareto front of the objective values, which is the set of nondominated points.
        """
        hv = []
        pf = None
        for i in range(iters):
            # Compute the hypervolume for the current set of points (up to train_shape + i)
            bd1 = FastNondominatedPartitioning(ref_point=ref_point, Y=-1*torch.tensor(Y[:train_shape + i, :]))
            hv.append(bd1.compute_hypervolume())
            pf = bd1.pareto_Y  # Store the current Pareto front

        return hv, pf



import torch
from typing import Optional, Callable
from torch import Tensor
from scipy.spatial.distance import cdist
from botorch.utils.transforms import normalize
from botorch.models.model_list_gp_regression import ModelListGP
from botorch.acquisition.objective import GenericMCObjective
from botorch.utils.sampling import draw_sobol_samples, sample_simplex
from botorch.utils.multi_objective.scalarization import get_chebyshev_scalarization
from botorch.sampling.normal import SobolQMCNormalSampler
from botorch.acquisition.multi_objective.parego import qLogNParEGO
from botorch.acquisition.logei import qLogNoisyExpectedImprovement
from botorch.acquisition.multi_objective.logei import qLogExpectedHypervolumeImprovement
from botorch.acquisition.multi_objective.utils import (
    sample_optimal_points,
    random_search_optimizer,
    compute_sample_box_decomposition,
)
from botorch.acquisition.multi_objective.predictive_entropy_search import (
    qMultiObjectivePredictiveEntropySearch,
)
from botorch.acquisition.multi_objective.max_value_entropy_search import (
    qLowerBoundMultiObjectiveMaxValueEntropySearch,
)
from botorch.acquisition.multi_objective.joint_entropy_search import (
    qLowerBoundMultiObjectiveJointEntropySearch,
)
from botorch.optim.optimize import optimize_acqf, optimize_acqf_list
from botorch.utils.multi_objective.box_decompositions import FastNondominatedPartitioning
from qpots.utils.utils import unstandardize, select_candidates
from qpots.utils.pymoo_problem import PyMooFunction, nsga2
from qpots.function import Function
from qpots.tsemo_runner import TSEMORunner


class Acquisition:
    """
    A class providing various acquisition functions and methods for multi-objective optimization.
    """

    def __init__(
        self,
        func: Function,
        gps: ModelListGP,
        cons: Optional[Callable] = None,
        device: torch.device = torch.device("cpu"),
        q: int = 1,
        NUM_RESTARTS: int = 10,
        RAW_SAMPLES: int = 512
    ) -> None:
        """
        Initialize the multi-objective acquisition class.

        Parameters:
            func (Callable): The test function being optimized.
            gps (ModelListGP): Gaussian Process models.
            cons (Optional[Callable]): A vector-valued function of inequality constraints.
            device (torch.device): Device for computation (CPU/GPU).
            q (int, optional): Number of points to sample per iteration. Defaults to 1.
            NUM_RESTARTS (int, optional): Number of restarts for acquisition optimization.
            RAW_SAMPLES (int, optional): Number of raw samples for acquisition optimization.
        """
        self.func = func
        self.gps = gps
        self.cons = cons
        self.device = device
        self.q = q
        self.NUM_RESTARTS = NUM_RESTARTS
        self.RAW_SAMPLES = RAW_SAMPLES
        self.nobj = gps.nobj
        self.ncons = gps.ncons

    def nystrom_approx(
        self,
        x: Tensor,
        gps: ModelListGP,
        m: int,
        pareto_set: Optional[Tensor] = None,
        col_choice: str = "pareto",
        seed_iter: int = 1,
        reg: float = 1e-6,
    ) -> Tensor:
        """
        Perform Thompson sampling using the Nystrom approximation for GP models.

        Parameters:
            x (Tensor): Input points.
            gps (ModelListGP): A list of GP models.
            m (int): Number of landmarks for the Nystrom approximation.
            pareto_set (Optional[Tensor]): Pareto optimal points. Required if `col_choice` is 'pareto'.
            col_choice (str, optional): Column selection method ('pareto' or 'random'). Defaults to 'pareto'.
            seed_iter (int, optional): Seed iteration for randomness. Defaults to 1.
            reg (float, optional): Regularization constant for matrix inversion. Defaults to 1e-6.

        Returns:
            Tensor: Posterior samples from the GP models.
        """
        torch.manual_seed(1024 + seed_iter)
        samples_list = []

        for gp_model in gps.models:
            posterior = gp_model.posterior(normalize(x.to(self.device), gps.bounds.to(self.device)))
            mean = posterior.mean.to(self.device)
            covariance = posterior.mvn.covariance_matrix.to(self.device)

            if col_choice == "random":
                indices = torch.randperm(covariance.shape[-1])[:m]
            elif col_choice == "pareto":
                if pareto_set is None:
                    raise ValueError("Pareto set is required for 'pareto' column choice.")
                D = cdist(pareto_set.cpu(), x.cpu().numpy())
                cand_indices = D.argmin(axis=-1)
                indices = torch.unique(torch.tensor(cand_indices).argsort()[:m]).to(self.device)
                if len(indices) < m:
                    extra_indices = torch.randperm(covariance.shape[-1])[:m - len(indices)].to(self.device)
                    indices = torch.cat((indices, extra_indices)).to(self.device)
            else:
                raise NotImplementedError(f"Column choice '{col_choice}' is not implemented.")

            K_nm = covariance[:, indices]
            apprx_covariance = covariance[indices][:, indices]
            while True:
                try:
                    L_mm_cov = torch.linalg.cholesky(
                        apprx_covariance + reg * torch.eye(m).to(self.device)
                    ).to(self.device)
                    break
                except torch.linalg.LinAlgError:
                    reg *= 10

            z = torch.rand(m, dtype=torch.float64).to(self.device)
            sample = mean + K_nm @ (L_mm_cov @ z).reshape(-1, 1)
            samples_list.append(sample.detach())

        Ys_ = unstandardize(torch.cat(samples_list, -1), gps.train_y.to(self.device))

        if self.ncons > 0:
            ind_feasible = Ys_[..., -self.ncons :] <= 0
            Ys_[~ind_feasible.squeeze(), : self.nobj] = -1e12  # Arbitrary low value for infeasible points
            Ys = Ys_[..., : self.nobj]
        else:
            Ys = Ys_

        return -Ys

    def gp_posterior(self, x: Tensor, gps: ModelListGP, seed_iter: int = 1) -> Tensor:
        """
        Compute posterior samples for PyMoo optimization.

        Parameters:
            x (Tensor): Input points.
            gps (ModelListGP): Gaussian Process models.
            seed_iter (int, optional): Iteration index for seeding randomness. Defaults to 1.

        Returns:
            Tensor: Posterior samples.
        """
        torch.manual_seed(1024 + seed_iter)
        Ys_ = [
            model.posterior(normalize(x, gps.bounds)).sample().reshape(-1, 1) for model in gps.models
        ]
        Ys_ = unstandardize(torch.cat(Ys_, -1), gps.train_y.to(self.device))

        if self.ncons > 0:
            ind_feasible = (Ys_[..., -self.ncons :] >= 0).all(dim=-1)
            Ys_[~ind_feasible.squeeze(), : self.nobj] = -1e12
            Ys = Ys_[..., : self.nobj]
        else:
            Ys = Ys_

        return -Ys

    
    def qpots(
        self,
        bounds: Tensor,
        iteration: int,
        **kwargs,
    ) -> Tensor:
        """
        Perform Pareto Optimal Thompson Sampling (QPOTS).

        Parameters:
            bounds (Tensor): The bounds for the optimization problem.
            iteration (int): Current iteration index for seeding randomness.
            **kwargs: Additional arguments for customization.

        Returns:
            Tensor: Selected candidates.
        """
        def track_pareto(res):
            pareto_set[0] = res.opt.get("X")

        pareto_set = [None]

        if kwargs["nystrom"] == 1:
            gp_posterior_approx_ = lambda x: self.nystrom_approx(
                x.to(self.device),
                self.gps,
                int(0.2 * kwargs["iters"]),
                normalize(torch.tensor(pareto_set[0]).to(self.device), bounds).cpu().numpy()
                if pareto_set[0] is not None
                else torch.zeros_like(x),
                col_choice=kwargs["nychoice"],
            )
            pymoo_func_gp = PyMooFunction(
                gp_posterior_approx_,
                n_var=kwargs["dim"],
                n_obj=self.nobj,
                xl=bounds[0].detach().cpu().numpy(),
                xu=bounds[1].detach().cpu().numpy(),
            )
            res = nsga2(
                pymoo_func_gp,
                ngen=kwargs["ngen"],
                pop_size=100 * kwargs["dim"],
                seed=2430,
                callback=track_pareto,
            )
        else:
            gp_posterior_ = lambda x: self.gp_posterior(
                x.to(self.device), self.gps, seed_iter=iteration
            )
            pymoo_func_gp = PyMooFunction(
                gp_posterior_,
                n_var=kwargs["dim"],
                n_obj=self.nobj,
                xl=bounds[0].detach().cpu().numpy(),
                xu=bounds[1].detach().cpu().numpy(),
            )
            res = nsga2(
                pymoo_func_gp,
                ngen=kwargs["ngen"],
                pop_size=100 * kwargs["dim"],
                seed=2430,
            )
        
        selected_candidates = select_candidates(
            self.gps, res.X, self.device, q=kwargs["q"], seed=2043
        )
        
        return normalize(selected_candidates, bounds)

    def qlogei(self, ref_point: Tensor = torch.tensor([0.0, 0.0])) -> Tensor:
        """
        Optimize the qLogEI acquisition function and return new candidates.

        Parameters:
            ref_point (Tensor, optional): Reference point for hypervolume calculation. Defaults to [0.0, 0.0].

        Returns:
            Tensor: New candidate points based on qLogEI optimization.
        """
        standard_bounds = torch.zeros(2, self.func.dim, device=self.device)
        standard_bounds[1] = 1
        train_y = self.gps.train_y.to(self.device)
        partitioning = FastNondominatedPartitioning(
            ref_point=ref_point.to(self.device), Y=train_y[..., : self.nobj]
        )
        model = ModelListGP(*self.gps.models).to(self.device)
        acq_func = qLogExpectedHypervolumeImprovement(
            model=model,
            ref_point=ref_point.to(self.device),
            partitioning=partitioning,
            constraints=[lambda Z: Z[..., -self.gps.ncons]],
        )

        while True:
            try:
                new_x, _ = optimize_acqf(
                    acq_function=acq_func,
                    bounds=standard_bounds,
                    q=self.q,
                    num_restarts=self.NUM_RESTARTS,
                    raw_samples=self.RAW_SAMPLES,
                    options={"batch_limit": 3, "maxiter": 1000},
                    sequential=True,
                )
                break
            except RuntimeError:
                pass
        return new_x.to(self.device)

    def parego(self) -> Tensor:
        """
        Perform qParEGO optimization using random weights for scalarization.

        Returns:
            Tensor: New candidate points based on qParEGO optimization.
        """
        standard_bounds = torch.tensor(
            [[0.0] * self.func.dim, [1.0] * self.func.dim],
            dtype=torch.double,
        ).to(self.device) # normalized bounds
        train_x = self.gps.train_x.to(self.device)
        train_y = self.gps.train_y.to(self.device)
        sampler = SobolQMCNormalSampler(sample_shape=torch.Size([128]))
        model = ModelListGP(*self.gps.models).to(self.device)
        pred = unstandardize(model.posterior(train_x).mean, train_y)

        acq_func_list = []
        for _ in range(self.q):
            weights = sample_simplex(self.func.nobj + self.gps.ncons).squeeze()
            objective = GenericMCObjective(get_chebyshev_scalarization(weights=weights, Y=pred))
            acq_func = qLogNoisyExpectedImprovement(
                model=model,
                objective=objective,
                X_baseline=train_x,
                sampler=sampler,
                prune_baseline=True,
                constraints=[lambda z: z[..., -self.gps.ncons]],
            )
            acq_func_list.append(acq_func)

        new_x, _ = optimize_acqf_list(
            acq_function_list=acq_func_list,
            bounds=standard_bounds,
            num_restarts=self.NUM_RESTARTS,
            raw_samples=self.RAW_SAMPLES,
            options={"batch_limit": 3, "maxiter": 1000},
        )
        return new_x.to(self.device)
    
    def qlogparego(self) -> Tensor:
        """
        Perform qParEGO optimization using qNParEGO from BoTorch.

        Returns:
            Tensor: New candidate points based on qParEGO optimization.
        """
        # Standardize bounds for optimization
        standard_bounds = torch.tensor(
            [[0.0] * self.func.dim, [1.0] * self.func.dim],
            dtype=torch.double,
        ).to(self.device) # normalized bounds


        train_x = self.gps.train_x.to(self.device)
        sampler = SobolQMCNormalSampler(sample_shape=torch.Size([128]))
        model = ModelListGP(*self.gps.models).to(self.device)

        # Use qNParEGO for acquisition function
        acq_func = qLogNParEGO(
            model=model,
            X_baseline=train_x,
            sampler=sampler,
            constraints=[lambda z: z[..., -self.gps.ncons]],
            prune_baseline=True, # This is recommended to improve performance
        )

        # Optimize the acquisition function
        new_x, _ = optimize_acqf(
            acq_function=acq_func,
            bounds=standard_bounds,
            q=self.q,
            num_restarts=self.NUM_RESTARTS,
            raw_samples=self.RAW_SAMPLES,
            options={"batch_limit": 3, "maxiter": 1000},
        )

        return new_x.to(self.device)

    def pesmo(self) -> Tensor:
        """
        Perform Predictive Entropy Search for Multi-Objective optimization (PESMO).

        Returns:
            Tensor: Candidate points generated by PESMO.
        """
        dim = self.func.dim
        bounds = torch.row_stack([torch.zeros(dim), torch.ones(dim)]).to(self.device)
        model = ModelListGP(*self.gps.models).to(self.device)

        ps, _ = sample_optimal_points(
            model=model,
            bounds=bounds.double(),
            num_samples=11,
            num_points=3,
            optimizer=random_search_optimizer,
            optimizer_kwargs={"pop_size": 10000, "max_tries": 45},
        )
        acqf = qMultiObjectivePredictiveEntropySearch(model=model, pareto_sets=ps)

        candidates, _ = optimize_acqf(
            acq_function=acqf,
            bounds=bounds.double(),
            q=self.q,
            num_restarts=5,
            raw_samples=512,
            options={"with_grad": False}
        )
        return candidates.to(self.device)

    def mesmo(self) -> Tensor:
        """
        Perform Multi-Objective Max-Value Entropy Search (MESMO).

        Returns:
            Tensor: Candidate points generated by MESMO.
        """
        dim = self.func.dim
        bounds = torch.row_stack([torch.zeros(dim), torch.ones(dim)]).to(self.device)
        model = ModelListGP(*self.gps.models).to(self.device)

        ps, pf = sample_optimal_points(
            model=model,
            bounds=bounds.double(),
            num_samples=10,
            num_points=10,
            optimizer=random_search_optimizer,
            optimizer_kwargs={"pop_size": 10000, "max_tries": 45},
        )
        hypercell_bounds = compute_sample_box_decomposition(pf)

        acqf = qLowerBoundMultiObjectiveMaxValueEntropySearch(
            model=model,
            hypercell_bounds=hypercell_bounds,
            estimation_type="LB",
        )
        candidates, _ = optimize_acqf(
            acq_function=acqf,
            bounds=bounds.double(),
            q=self.q,
            num_restarts=8,
            raw_samples=512,
            sequential=True,
        )
        return candidates.to(self.device)

    def jesmo(self) -> Tensor:
        """
        Perform Joint Entropy Search for Multi-Objective optimization (JESMO).

        Returns:
            Tensor: Candidate points generated by JESMO.
        """
        dim = self.func.dim
        bounds = torch.row_stack([torch.zeros(dim), torch.ones(dim)]).to(self.device)
        model = ModelListGP(*self.gps.models).to(self.device)

        ps, pf = sample_optimal_points(
            model=model,
            bounds=bounds.double(),
            num_samples=10,
            num_points=10,
            optimizer=random_search_optimizer,
            optimizer_kwargs={"pop_size": 10000, "max_tries": 45},
        )
        hypercell_bounds = compute_sample_box_decomposition(pf)

        acqf = qLowerBoundMultiObjectiveJointEntropySearch(
            model=model,
            pareto_sets=ps,
            pareto_fronts=pf,
            hypercell_bounds=hypercell_bounds,
            estimation_type="LB",
        )
        candidates, _ = optimize_acqf(
            acq_function=acqf,
            bounds=bounds.double(),
            q=self.q,
            num_restarts=8,
            raw_samples=512,
            sequential=True,
        )
        return candidates.to(self.device)

    def sobol(self) -> Tensor:
        """
        Generate random samples.

        Returns:
            Tensor: Randomly generated candidate points.
        """
        standard_bounds = torch.row_stack(
            [torch.zeros(self.func.dim), torch.ones(self.func.dim)]
        )
        return draw_sobol_samples(
            bounds=standard_bounds, n=1, q=self.q
        ).squeeze(1).to(self.device)
    
    def tsemo(self, save_dir, iters, ref_point, train_shape, rep=0):
        ts = TSEMORunner(self.func.name, 
                         self.gps.train_x, 
                         self.gps.train_y, 
                         lb=[0.0]*self.gps.train_x.shape[1], 
                         ub=[1.0]*self.gps.train_x.shape[1], 
                         iters=iters, 
                         batch_number=self.q)
        x, y, times = ts.tsemo_run(save_dir, rep)
        hv = ts.tsemo_hypervolume(y, ref_point, train_shape, iters)
        return x, y, times, hv

        
